/*
 * Copyright (c) 2005 Mellanox Technologies. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 * $Id: bench_test.c 833 2006-01-02 12:06:55Z dotanb $ 
 * 
 */

#include <stdlib.h>
#include <infiniband/verbs.h>
#include <vl.h>

#define TEST_NAME "alloc_mcast"

#define MCAST_LID 0xc000

struct VL_usage_descriptor_t usage_descriptor[] = {
    {
        'h', "help", "",
        "Print this message and exit",
        #define HELP_CMD_CASE                      0
        HELP_CMD_CASE
    },
    
    {
        'w', "wait", "",
        "Wait (getchar) before Cleanup",
        #define WAIT_CMD_CASE                      1
        WAIT_CMD_CASE
    },

    {
        'd', "device", "DEVICE_NAME",
        "Device to use. Default: mthca0",
        #define HCA_N_CMD_CASE                     2
        HCA_N_CMD_CASE
    },

    {
        'm', "num_of_mcast", "NUM_OF_MCAST",
        "Number of IB multicast groups to allocate (default 1)",
        #define NUM_OF_MCAST_CMD_CASE				3
        NUM_OF_MCAST_CMD_CASE
    },

    {
        's', "sleep", "SECONDS",
        "Time to sleep (in seconds) before detaching from multicast groups",
        #define SLEEP_BEFORE_CLEAN_CMD_CASE			4
        SLEEP_BEFORE_CLEAN_CMD_CASE
    }

};

struct config_t {
    int                     wait_cleanup;
    const char*             dev_name;
    int						num_of_mcast;
	int						sleep_time;
};

struct resources {
    struct ibv_device		**dev_list;
	struct ibv_context      *ib_ctx;
	struct ibv_pd		    *pd;
    struct ibv_cq           *cq;
    struct ibv_qp		    **qp_arr;
    struct ibv_device_attr  device_attr;
    int			            cq_size;
};

struct config_t config = {
    0,              /* wait cleanup */
    "mthca0",       /* dev_name */
    1,           	/* num_of_mcast */
	60            	/* sleep_time */
};

/*****************************************************************************
* Function: process_arg
*****************************************************************************/
static int process_arg(
    IN              const int opt_index,
    IN              const char* equ_ptr,
    IN              const int arr_size,
    IN              const struct VL_usage_descriptor_t* usage_desc_arr)
{
        /* process argument */
	switch (usage_descriptor[opt_index].case_code) {
	case HELP_CMD_CASE:
		VL_usage(1, arr_size, usage_desc_arr);
		return 1;

	case WAIT_CMD_CASE:
		config.wait_cleanup = 1;
		break;

	case HCA_N_CMD_CASE:
        config.dev_name = equ_ptr;
		break;

	case NUM_OF_MCAST_CMD_CASE:
		config.num_of_mcast = strtoul( equ_ptr, 0, 0 );
		break;

	case SLEEP_BEFORE_CLEAN_CMD_CASE:
		config.sleep_time = strtoul( equ_ptr, 0, 0 );
		break;

	default:
		VL_MISC_ERR(("unknown parameter is the switch %s\n",equ_ptr));
		exit(2);
	}

	return 0;
}

/*****************************************************************************
* Function: resources_init
*****************************************************************************/
static void resources_init(
	IN	struct resources *res)
{
	res->dev_list = NULL;
	res->ib_ctx = NULL;
    res->cq = NULL;
    res->pd = NULL;
    res->qp_arr = NULL;
}

/*****************************************************************************
* Function: set_start_gid
*****************************************************************************/
static void set_start_gid(
	union ibv_gid *gid_p)
{
	memset(gid_p, 0, sizeof(union ibv_gid));

	gid_p->raw[0] = 255;
	gid_p->raw[1] = 1;
	gid_p->raw[9] = 2;
	gid_p->raw[10] = 201;
	gid_p->raw[11] = 0;
	gid_p->raw[12] = 1;
	gid_p->raw[13] = 0;
	gid_p->raw[14] = 208;
	gid_p->raw[15] = 0;
}

/*****************************************************************************
* Function: increment_gid
*****************************************************************************/
static void increment_gid(
	union ibv_gid *gid_p)
{
	int i = 16;


	do {
		i --;
		gid_p->raw[i] ++;

	} while (gid_p->raw[i] == 0);
}

/*****************************************************************************
* Function: print_gid
*****************************************************************************/
void print_gid(
	IN				int idx,
    IN              const union ibv_gid *gid)
{
    VL_MISC_TRACE(("GID[%d] %u.%u.%u.%u.%u.%u.%u.%u.%u.%u.%u.%u.%u.%u.%u.%u",
				   idx,
                   gid->raw[0], gid->raw[1], gid->raw[2], gid->raw[3],
                   gid->raw[4], gid->raw[5], gid->raw[6], gid->raw[7],
                   gid->raw[8], gid->raw[9], gid->raw[10], gid->raw[11],
                   gid->raw[12], gid->raw[13], gid->raw[14], gid->raw[15]));
}

/*****************************************************************************
* Function: resources_create
*****************************************************************************/
static int resources_create(
	IN	struct resources *res)
{
    struct ibv_device *ib_dev = NULL;
    size_t size;
    int i;
    int cq_size;
	int num_devices;


    VL_HCA_TRACE(("Searching for IB devices in host"));

    /* get device names in the system */
    res->dev_list = ibv_get_device_list(&num_devices);
    if (!res->dev_list) {
        VL_HCA_ERR(("failed to get IB devices list"));
        return 1;
    }

    for (i = 0; i < num_devices; i ++) {
        if (!strcmp(ibv_get_device_name(res->dev_list[i]), config.dev_name)) {
			ib_dev = res->dev_list[i];
            break;
        }
    }

    if (!ib_dev) {
        VL_HCA_ERR(("IB device %s wasn't found", config.dev_name));
        return 1;
    }

    VL_HCA_TRACE(("Device %s was found", config.dev_name));

    /* get device handle */
    res->ib_ctx = ibv_open_device(ib_dev);
    if (!res->ib_ctx) {
        VL_HCA_ERR(("failed to open device %s", config.dev_name));
        return 1;
    }

	res->pd = ibv_alloc_pd(res->ib_ctx);
    if (!res->pd) {
        VL_HCA_ERR(("ibv_alloc_pd failed"));
        return 1;
    }

    cq_size = 1;
    res->cq = ibv_create_cq(res->ib_ctx, cq_size, NULL, NULL, 0);
    if (!res->cq) {
        VL_DATA_ERR(("failed to create CQ with %u entries", cq_size));
        return 1;
    }
    VL_DATA_TRACE(("CQ was created with %u CQEs", cq_size));

	size = sizeof(struct ibv_qp *) * config.num_of_mcast;
	res->qp_arr = VL_MALLOC(size, struct ibv_qp *);
	if (!res->qp_arr) {
        VL_MEM_ERR(("failed to allocate %Zu bytes for memory array", size));
        return 1;
	}

	for (i = 0; i < config.num_of_mcast; i ++)
		res->qp_arr[i] = NULL;

	for (i = 0; i < config.num_of_mcast; i ++) {
		struct ibv_qp_init_attr attr = {
			.send_cq = res->cq,
			.recv_cq = res->cq,
			.cap     = {
				.max_send_wr  = 1,
				.max_recv_wr  = 1,
				.max_send_sge = 1,
				.max_recv_sge = 1
			},
			.qp_type = IBV_QPT_UD,
			.sq_sig_all = 1
		};

		res->qp_arr[i] = ibv_create_qp(res->pd, &attr);
        if (!res->qp_arr[i]) {
            VL_DATA_ERR(("failed to create QP[%d]", i));
            return 1;
        }
		VL_DATA_TRACE(("QP[%d] was created, QP number=0x%x", i, res->qp_arr[i]->qp_num));
	}

    return 0;
}

/*****************************************************************************
* Function: resources_destroy
*****************************************************************************/
static int resources_destroy(
	IN	struct resources *res)
{
    int test_result = 0;
	int i;
    

	if (res->qp_arr) {
		for (i = 0; i < config.num_of_mcast; i ++) {
			if (res->qp_arr[i]) {
				if (ibv_destroy_qp(res->qp_arr[i])) {
					VL_DATA_ERR(("failed to destroy QP[%d]", i));
					test_result = 1;
				}
			}
		}

		VL_FREE(res->qp_arr);
	}

	if (res->cq) {
		if (ibv_destroy_cq(res->cq)) {
            VL_DATA_ERR(("ibv_destroy_cq of CQ failed"));
            test_result = 1;
        }
	}

	if (res->pd) {
		if (ibv_dealloc_pd(res->pd)) {
            VL_HCA_ERR(("ibv_dealloc_pd failed"));
            test_result = 1;
        }
	}
        
	if (res->ib_ctx) {
		if (ibv_close_device(res->ib_ctx)) {
            VL_HCA_ERR(("ibv_close_device failed"));
            test_result = 1;
        }
	}
 
	if (res->dev_list)
		ibv_free_device_list(res->dev_list);

    return test_result;
}
/*****************************************************************************
* Function: print_config
*****************************************************************************/
static void print_config(void)
{
	VL_MISC_TRACE((" %s configuration report", TEST_NAME));
	VL_MISC_TRACE((" ------------------------------------------------"));
	VL_MISC_TRACE((" Current pid                  : "VL_PID_FMT, VL_getpid()));
	VL_MISC_TRACE((" Device name                  : \"%s\"", config.dev_name));
    VL_MISC_TRACE((" Number of IB multicast groups: %u", config.num_of_mcast));
	VL_MISC_TRACE((" Wait before cleanup          : %s", ((config.wait_cleanup)?"Yes":"No")));
	VL_MISC_TRACE((" Time to sleep before cleanup : %u", config.sleep_time));
    VL_MISC_TRACE((" ------------------------------------------------\n"));
}

/*****************************************************************************
******************************************************************************
* Function: main
******************************************************************************
*****************************************************************************/
int main(int argc, char *argv[])
{
    struct resources res;
	union ibv_gid gid;
    int test_result = 1;
	int rc;
	int i;
    
	
	rc = VL_parse_argv(argc, argv, (sizeof(usage_descriptor)/sizeof(struct VL_usage_descriptor_t)),
                          usage_descriptor, process_arg);
	if (rc != 0)
		return rc;

	resources_init(&res);

    print_config();

    if (resources_create(&res))
        goto cleanup;

    /* attach all the QPs to multicast groups */
	set_start_gid(&gid);
	for (i = 0; i < config.num_of_mcast; i ++) {
		rc = ibv_attach_mcast(res.qp_arr[i], &gid, MCAST_LID);
		if (rc) {
			VL_DATA_ERR(("failed to attach QP[%u] to multicast group", i));
			goto cleanup;
		}

		print_gid(i, &gid);
        increment_gid(&gid);
	}

	VL_DATA_TRACE(("%u IB multicast groups were created (each QP in one group)", config.num_of_mcast));


	VL_MISC_TRACE(("Going to sleep for %u seconds", config.sleep_time));
    VL_sleep(config.sleep_time);

	test_result = 0;

cleanup:
	if (config.wait_cleanup)
		VL_keypress_wait();

    /* detach all the QPs to multicast groups */
	if (res.qp_arr) {
        set_start_gid(&gid);

        for (i = 0; i < config.num_of_mcast; i ++) {
            rc = ibv_detach_mcast(res.qp_arr[i], &gid, MCAST_LID);
            if (rc) {
                VL_DATA_ERR(("failed to detach QP[%u] from multicast group", i));
                test_result = 1;
			}

			print_gid(i, &gid);
			increment_gid(&gid);
		}
	}

	VL_DATA_TRACE(("%u IB multicast groups were detached (each QP in one group)", config.num_of_mcast));


    if (resources_destroy(&res))
        test_result = 1;

    VL_print_test_status(test_result);

    return test_result;
}

